
#include <clutter/clutter.h>
#include <clutter-gst/clutter-gst.h>


typedef enum
{
	PAD_BLOCKING_None,
	PAD_BLOCKING_Block,
	PAD_BLOCKING_UnBlock

} PadBlocking;


typedef struct _MyMediaData
{
	GstBus			*bus;

	GstElement		*pipeline;
	GstElement		*src;
	//GstElement	*playbin2;
	GstElement		*decoder;		// decodebin2
	gboolean		decoder_has_audio;
	gboolean		decoder_has_video;
	GstPad			*decoder_audio_pad;
	GstPad			*decoder_video_pad;

	GstElement		*audio_tee;

	GstElement		*AudioBin;
	GstElement		*audio_queue;
	GstElement		*audio_convert;
	GstElement		*audio_resample;
	GstElement		*audio_sink;
	GstElement		*visualizer_queue;
	GstElement		*visualizer_convert;
	GstElement		*visualizer_visual;
	GstElement		*visualizer_color_space;
	//GstElement	*video_sink;	// ,,,

	GstElement		*VideoBin;
	GstElement		*video_queue;
	GstElement		*video_color_space;
	//GstElement	*video_sink;	// using clutter_video_sink instead



	ClutterActor	*clutter_stage;
	ClutterActor	*clutter_video_texture;
	GstElement		*clutter_video_sink;

} MyMediaData;





void
size_change (ClutterTexture *texture,
             gfloat          width,
             gfloat          height,
             gpointer        user_data)
{
  ClutterActor *stage;
  gfloat new_x, new_y, new_width, new_height;
  gfloat stage_width, stage_height;

  stage = clutter_actor_get_stage (CLUTTER_ACTOR (texture));
  if (stage == NULL)
    return;

  clutter_actor_get_size (stage, &stage_width, &stage_height);

  new_height = (height * stage_width) / width;
  if (new_height <= stage_height)
    {
      new_width = stage_width;

      new_x = 0;
      new_y = (stage_height - new_height) / 2;
    }
  else
    {
      new_width  = (width * stage_height) / height;
      new_height = stage_height;

      new_x = (stage_width - new_width) / 2;
      new_y = 0;
    }

  clutter_actor_set_position (CLUTTER_ACTOR (texture), new_x, new_y);
  clutter_actor_set_size (CLUTTER_ACTOR (texture), new_width, new_height);
}


static gint
AudioBin_Create( MyMediaData *data )
{

	data->AudioBin = gst_bin_new( NULL );
	data->audio_tee = gst_element_factory_make( "tee", NULL );

	gst_bin_add_many( GST_BIN( data->AudioBin ), data->audio_tee, NULL );
	gst_bin_add( GST_BIN( data->pipeline ), data->AudioBin );


	return 0;
}


static gint
AudioBin_Link_to_Decoder( MyMediaData *data )
{

	GstPad *audio_tee_pad = gst_element_get_pad( data->audio_tee, "sink" );
	gst_element_add_pad( data->AudioBin, gst_ghost_pad_new( "sink", audio_tee_pad ) );
	gst_object_unref( audio_tee_pad );


	GstPad *audio_bin_pad = gst_element_get_pad( data->AudioBin, "sink" );
	gst_pad_link( data->decoder_audio_pad, audio_bin_pad );
	gst_object_unref( audio_bin_pad );
	//gst_object_unref( data->decoder_audio_pad );		// does this need to be unReffed?


    gst_element_set_state( data->AudioBin, GST_STATE_PAUSED );


	return 0;
}


static gint
VideoBin_Create_and_Add_Elements( MyMediaData *data )
{

    data->VideoBin = gst_bin_new( NULL );

    data->video_queue = gst_element_factory_make( "queue", NULL );
	g_object_set( data->video_queue, "max-size-time", (guint64) 3 * GST_SECOND, "max-size-bytes", (guint64) 0, "max-size-buffers", (guint64) 0, NULL);

	data->video_color_space = gst_element_factory_make("ffmpegcolorspace", NULL);

	gst_bin_add_many( GST_BIN( data->VideoBin ), data->video_queue, data->video_color_space, data->clutter_video_sink, NULL );
	gst_element_link_many( data->video_queue, data->video_color_space, data->clutter_video_sink, NULL );

    gst_bin_add( GST_BIN( data->pipeline ), data->VideoBin );


	return 0;
}


static gint
VideoBin_Link_to_Decoder( MyMediaData *data )
{

	GstPad *video_queue_pad = gst_element_get_pad( data->video_queue, "sink" );
	GstPad *video_queue_pad_ghost = gst_ghost_pad_new( "sink", video_queue_pad );
	gst_object_unref( video_queue_pad );

    gst_element_add_pad( data->VideoBin, video_queue_pad_ghost );
    gst_pad_link( data->decoder_video_pad, video_queue_pad_ghost );

    gst_element_set_state( data->VideoBin, GST_STATE_PAUSED );


	return 0;
}


static gint
AudioBin_Add_Elements( MyMediaData *data )
{

	data->audio_queue = gst_element_factory_make("queue", NULL);
	g_object_set( data->audio_queue, "max-size-time", (guint64) 3 * GST_SECOND, "max-size-bytes", (guint64) 0, "max-size-buffers", (guint64) 0, NULL);

	data->audio_convert = gst_element_factory_make("audioconvert", NULL);
	data->audio_resample = gst_element_factory_make("audioresample", NULL);

	data->audio_sink = gst_element_factory_make("alsasink", NULL);

	gst_bin_add_many( GST_BIN( data->AudioBin ), data->audio_queue, data->audio_convert, data->audio_resample, data->audio_sink, NULL );
	gst_element_link_many( data->audio_queue, data->audio_convert, data->audio_resample, data->audio_sink, NULL );

	GstPad *audio_queue_pad = gst_element_get_pad( data->audio_queue, "sink" );
	GstPad *audio_tee_pad = gst_element_get_request_pad( data->audio_tee, "src%d" );

	gst_pad_link( audio_tee_pad, audio_queue_pad );

	gst_object_unref( audio_tee_pad );
	gst_object_unref( audio_queue_pad );


	return 0;
}


static gint
VisualizerSegment_Add_Elements( MyMediaData *data )
{

	data->visualizer_queue = gst_element_factory_make( "queue", NULL );
	g_object_set( data->visualizer_queue, "max-size-time", (guint64) 3 * GST_SECOND, "max-size-bytes", (guint64) 0, "max-size-buffers", (guint64) 0, NULL );

	data->visualizer_convert = gst_element_factory_make( "audioconvert", NULL );
	data->visualizer_visual = gst_element_factory_make( "goom", NULL );
	data->visualizer_color_space = gst_element_factory_make( "ffmpegcolorspace", NULL );
	//sink = gst_element_factory_make( "ximagesink", NULL );	// using clutter_video_sink instead

	gst_bin_add_many( GST_BIN( data->AudioBin ), data->visualizer_queue, data->visualizer_convert, data->visualizer_visual, data->visualizer_color_space, data->clutter_video_sink, NULL );
	gst_element_link_many( data->visualizer_queue, data->visualizer_convert, data->visualizer_visual, data->visualizer_color_space, data->clutter_video_sink, NULL );

	GstPad *visualizer_queue_pad = gst_element_get_pad( data->visualizer_queue, "sink" );
	GstPad *audio_tee_pad = gst_element_get_request_pad( data->audio_tee, "src%d" );
	gst_pad_link( audio_tee_pad, visualizer_queue_pad );

	gst_object_unref( audio_tee_pad );
	gst_object_unref( visualizer_queue_pad );


	return 0;
}


static gboolean
on_Bus_Message( GstBus *bus, GstMessage *msg, gpointer data )
{
	gchar *debug;
	GError *err;

	switch(GST_MESSAGE_TYPE (msg))
	{
		case GST_MESSAGE_ASYNC_DONE:
			g_print ("ASYNC is Done\n");
			break;

		case GST_MESSAGE_EOS:
			g_print( "End-of-stream\n" );
			// errrr... do sumtin'!
			break;

		case GST_MESSAGE_ERROR:
			gst_message_parse_error( msg, &err, &debug );
			g_free( debug );

			g_print( "Error: %s\n", err->message );
			g_error_free( err );

			// MAYDAY!!!  MAN OVERBOARD!!!  All try to Save him!  Now, ALL OVERBOARD!!!  (8-(
			// errrr... do sumtin'!
			break;

		default:
			break;
	}

	return TRUE;
}


static void
on_DecoderPad_Blocked( GstPad *pad, gboolean blocked, gpointer user_data )
{
	gchar *name = gst_pad_get_name( pad );

	if( blocked )	g_debug( "Pad %s blocked successfully for %s", name, (gchar *) user_data );
	else			g_debug( "Pad %s unblocked successfully for %s", name, (gchar *) user_data );

	g_free( name );

	gst_object_unref( GST_OBJECT( pad ) );
}


static gint
DecoderPads_Set_Blocking( MyMediaData *data, PadBlocking blocking )
{
	gint			ret = -1;
    GstIterator		*it;
    gpointer		ptr_pad;
    gboolean		block_pads = ( blocking == PAD_BLOCKING_Block );

    it = gst_element_iterate_src_pads( data->decoder );

	while( gst_iterator_next( it, & ptr_pad ) == GST_ITERATOR_OK )
    {
        GstPad *pad = GST_PAD( ptr_pad );
        GstCaps *caps;
        gchar *str;

        g_print( "stream: %s   ", GST_OBJECT_NAME( pad ) );
        caps = gst_pad_get_caps( pad );
        str = gst_caps_to_string( caps );
        gst_caps_unref( caps );

        if( g_str_has_prefix( str, "video/" ) )
        {
        	data->decoder_has_video = TRUE;
        	gst_pad_set_blocked_async( pad, block_pads, on_DecoderPad_Blocked, NULL );
        	if( block_pads )	data->decoder_video_pad = pad;
        	//else        		gst_object_unref( data->decoder_video_pad );
        }
        else if( g_str_has_prefix( str, "audio/" ) )
        {
        	data->decoder_has_audio = TRUE;
        	gst_pad_set_blocked_async( pad, block_pads, on_DecoderPad_Blocked, NULL );
        	if( block_pads )	data->decoder_audio_pad = pad;
        	//else        		gst_object_unref( data->decoder_audio_pad );
        	//   assume only 1 audio pad! (:/
        }

        g_free( str );

        //if( block_pads == FALSE )
        //	gst_object_unref( pad );	// since data->decoder_audio_pad used these pads for the video/audio bins, we should let the pipeline free them!?!?
    }

    gst_iterator_free( it );


	return ret;
}


static gint
Wait_for_DecoderPads_Creation( MyMediaData *data )
{
	gint						ret = -1;


/*
	GstMessage *msg = NULL;
	msg = gst_bus_timed_pop_filtered( data->bus, 2 * GST_SECOND, GST_MESSAGE_ASYNC_DONE );
	//msg = gst_bus_timed_pop_filtered( data->bus, GST_CLOCK_TIME_NONE, GST_MESSAGE_ASYNC_DONE );
	if( msg )
	{
		if( GST_MESSAGE_TYPE( msg ) == GST_MESSAGE_ERROR )
		{
			g_printerr("Failed to send message GST_MESSAGE_ASYNC_DONE to the bus. Exiting.\n");
			goto EndOfFunction;
		}
		gst_message_unref( msg );
	}
	else
	{
		g_printerr("Failed to get Message from the BUS.\n");
		//
	}
	gst_object_unref( data->bus );
*/


	GstStateChangeReturn  res;
	res = gst_element_set_state( data->pipeline, GST_STATE_PAUSED );
	if (res == GST_STATE_CHANGE_FAILURE)
	{
		g_print( "Gstreamer ERROR: could not pause\n" );
		goto EndOfFunction;
	}

	res = gst_element_get_state( data->pipeline, NULL, NULL, GST_CLOCK_TIME_NONE );
	if( res != GST_STATE_CHANGE_SUCCESS )
	{
		g_print( "Gstreamer ERROR: could not complete pause\n" );
		goto EndOfFunction;
	}



	ret = 0;


EndOfFunction:
	return ret;
}


static gint
Pipeline_and_Decoder_Create( MyMediaData *data, char* mediafile )
{

	data->pipeline = gst_pipeline_new( "pipeline" );


	data->bus = gst_pipeline_get_bus( GST_PIPELINE( data->pipeline ) );
	gst_bus_add_watch( data->bus, on_Bus_Message, NULL );
	gst_object_unref( data->bus );


	data->src = gst_element_factory_make( "filesrc", "file-source" );
	g_object_set( G_OBJECT( data->src ), "location", mediafile, NULL );


	//data->decoder = gst_element_factory_make("playbin2", "decoder");
	data->decoder = gst_element_factory_make ("decodebin2", "decoder");


	//g_signal_connect( data->decoder, "new-decoded-pad", G_CALLBACK( on_decoder_pad_added ), NULL );


	gst_bin_add_many( GST_BIN( data->pipeline ), data->src, data->decoder, NULL );
	gst_element_link( data->src, data->decoder );


	return 0;
}


static gint
Pipeline_Destroy( MyMediaData *data )
{
	gst_element_set_state( data->pipeline, GST_STATE_NULL );
	gst_object_unref( GST_OBJECT( data->pipeline ) );

	return 0;
}


static gint
Clutter_Stage_Init( MyMediaData *data )
{
	data->clutter_stage = clutter_stage_get_default();
	clutter_actor_set_size( CLUTTER_ACTOR( data->clutter_stage ), 640, 360 );

	// We need to set certain props on the target texture currently for efficient/corrent playback onto the texture
	data->clutter_video_texture = g_object_new( CLUTTER_TYPE_TEXTURE, "sync-size", FALSE, "disable-slicing", TRUE, NULL );
	clutter_group_add( CLUTTER_GROUP( data->clutter_stage ), data->clutter_video_texture );
	clutter_actor_set_position( data->clutter_video_texture, 0 ,0 );
	g_signal_connect( CLUTTER_TEXTURE( data->clutter_video_texture ), "size-change", G_CALLBACK( size_change ), NULL );


	data->clutter_video_sink = clutter_gst_video_sink_new( CLUTTER_TEXTURE( data->clutter_video_texture ) );
	g_object_set( G_OBJECT( data->clutter_video_sink ), "qos", TRUE, "sync", TRUE, NULL );


	return 0;
}



int
main( int argc, char *argv[] )
{
	if (argc < 1)
	{
		g_error( "Usage: %s mediafile", argv[0] );
		goto EndOfFunction;
	}

	if( clutter_init( &argc, &argv ) != CLUTTER_INIT_SUCCESS )
		goto EndOfFunction;

	gst_init( &argc, &argv );


	MyMediaData data = {0};

	Clutter_Stage_Init( & data );

	Pipeline_and_Decoder_Create( & data, argv[1] );

	if( Wait_for_DecoderPads_Creation( & data ) == -1 )		goto EndOfFunction;

	DecoderPads_Set_Blocking( & data, PAD_BLOCKING_Block );		// Block

	// Setup the Bin PIPES and... SMOKE EM! :D
    if( data.decoder_has_video )
    {
    	if( data.decoder_has_audio )
    	{
    		AudioBin_Create( & data );

    		AudioBin_Add_Elements( & data );

    		AudioBin_Link_to_Decoder( & data );
    	}

    	VideoBin_Create_and_Add_Elements( & data );

    	VideoBin_Link_to_Decoder( & data );
    }
    else
    {
    	if( data.decoder_has_audio )
    	{
    		AudioBin_Create( & data );

    		AudioBin_Add_Elements( & data );

    		VisualizerSegment_Add_Elements( & data );

    		AudioBin_Link_to_Decoder( & data );
    	}
    }

    DecoderPads_Set_Blocking( & data, PAD_BLOCKING_UnBlock );		// Unblock


    gst_element_set_state( data.pipeline, GST_STATE_PLAYING );


	clutter_actor_show_all( data.clutter_stage );
	clutter_main();		  // main loop


EndOfFunction:
	Pipeline_Destroy( & data );
	return 0;
}


